from .app_feature import AppFeature
from.textual_control_feature import TextualControlFeature
from .textual_page import TextualPage


__all__ = [
    'AppFeature',
    'TextualControlFeature',
    'TextualPage',
]



import textual

from balderhub.textual.lib.scenario_features import AppFeature
from ...dut.calculator import CalculatorApp


class CalculatorAppFeature(AppFeature):

    def get_app(self) -> textual.app.App:
        return CalculatorApp()



import time
import balderhub.gui.lib.scenario_features
import balderhub.guicontrol.lib.scenario_features

import balderhub.textual.lib.utils.drivers


class TextualPage(balderhub.gui.lib.scenario_features.PageFeature):
    """
    Page object for textual application screens
    """

    guicontrol = balderhub.guicontrol.lib.scenario_features.GuiControlFeature()

    @property
    def driver(self) -> balderhub.textual.lib.utils.drivers.TextualDriverClass:
        """
        Returns the driver instance
        """
        return self.guicontrol.driver

    @property
    def applicable_on_screen_id(self) -> str:
        """
        Should return the screen id - is used to determine which screen is active at the moment
        """
        raise NotImplementedError

    def is_applicable(self):
        """
        Returns true if the current screen is the expected one (see :meth:`TextualPage.applicable_on_screen_id`)
        """
        # how should that be terminated?
        return self.driver.pilot_instance.app.screen.id == self.applicable_on_screen_id

    def wait_till_applicable(self, timeout=5):
        """
        Waits until the expected screen is active (default timeout is 5 seconds)
        """
        start_time = time.perf_counter()
        all_visibilities = []
        while True:
            current_screen_id = self.driver.pilot_instance.app.screen.id
            if current_screen_id == self.applicable_on_screen_id:
                # found it
                return
            if current_screen_id not in all_visibilities:
                all_visibilities.append(current_screen_id)
            time.sleep(0.1)
            if time.perf_counter() - start_time > timeout:
                raise TimeoutError(f'expected page with activity `{self.applicable_on_screen_id}` is not applicable '
                                   f'within timeout of {timeout} seconds (following activities were visible while '
                                   f'waiting: {all_visibilities}`')



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, List, Optional, TYPE_CHECKING, TypeVar

from textual.containers import VerticalScroll, HorizontalScroll, ScrollableContainer
from textual.dom import DOMNode
from textual.widget import Widget
from textual.widgets import Button, Label, Input, TextArea, Select, Digits

from balderhub.guicontrol.lib.utils.element_bridges.base_element_bridge import BaseElementBridge


from ..selector import Selector


if TYPE_CHECKING:
    from ..drivers import TextualDriverClass
    from .textual_fully_reidentifiable_element_bridge import TextualFullyReidentifiableElementBridge
    from .textual_partly_reidentifiable_element_bridge import TextualPartlyReidentifiableElementBridge
    from .textual_not_reidentifiable_element_bridge import TextualNotReidentifiableElementBridge

DOMNodeTypeT = TypeVar("DOMNodeTypeT", bound=DOMNode)


class TextualElementBridge(BaseElementBridge, ABC):
    """
    basic web-element-bridge implementation for Textual Elements
    """

    def __init__(self, driver: TextualDriverClass, parent: Optional[TextualElementBridge]):
        """
        Creates a new instance

        :param driver: the base driver class
        :param parent: the parent web element bridge (if this element has a parent element)
        """
        super().__init__(driver, parent)

    @property
    def driver(self) -> TextualDriverClass:
        """
        :return: returns the driver class, this bridge was created from
        """
        return self._driver

    @property
    def raw_element(self) -> DOMNodeTypeT:
        return super(self).raw_element

    @property
    def parent(self) -> Union[TextualElementBridge, None]:
        """
        :return: returns the defined parent web element bridge if a parent does exist
        """
        return self._parent

    def find_raw_element(self, selector: Selector) -> DOMNodeTypeT:
        """
        Method to find a specific raw web element by its selector that is a child of the current one

        :param selector: the selector specifying the element (relative to this one)
        :return: the raw web element (depending on the underlying framework)
        """
        result = self.raw_element.query_one(selector.to_textual_string())
        if result is None:
            raise ValueError(f'no such element with selector `{selector}`')
        return result

    def find_raw_elements(self, selector: Selector) -> List[DOMNodeTypeT]:
        """
        Method to find raw web elements matching the provided relative selector as a child of the current one

        :param selector: the selector specifying the elements (relative to this one)
        :return: the raw web element (depending on the underlying framework)
        """
        return list(self.raw_element.query(selector.to_textual_string()))

    @abstractmethod
    def find_bridge(
            self,
            selector: Selector
    ) -> Union[TextualFullyReidentifiableElementBridge, TextualPartlyReidentifiableElementBridge]:
        """
        Method to directly returning the bridge of a specific raw web element by its selector that is a child of the
        current one. In case that the element can be reidentified completely by the selector (f.e. because the selector
        is By.ID) the method returns a :class:`FullyReidentifiableElementBridge` object. Otherwise, the method returns
        a :class:`PartlyReidentifiableElementBridge` object.

        :param selector: the selector specifying the element (relative to this one)
        :return: the bridge object of the element specified by the selector
        """

    def find_bridges(self, selector: Selector) -> List[TextualNotReidentifiableElementBridge]:
        """
        Method to directly returning a list of bridge objects for raw web elements matching the provided relative
        selector as a child of the current one

        :param selector: the selector specifying the elements (relative to this one)
        :return: a list with all matching bridge objects (depending on the underlying framework)
        """
        # pylint: disable-next=import-outside-toplevel
        from .textual_not_reidentifiable_element_bridge import TextualNotReidentifiableElementBridge
        result = []
        for cur_elem in self.find_raw_elements(selector):
            result.append(TextualNotReidentifiableElementBridge(self.driver, cur_elem, parent=self))
        return result

    def exists(self) -> bool:
        """
        This method returns True if the element exists. An element exists if it is still part of the DOM.

        .. note::
            This does not mean, that it needs to be visible! Use the :meth:`BaseWebElementBridge.is_visible` method
            instead.`

        :return: returns True if the element does exist otherwise false.
        """
        return self.raw_element.display

    def is_visible(self) -> bool:
        """
        This method returns True if the element is visible.

        :return: returns True if the element is visible otherwise false.
        """
        return self.raw_element.visible

    def is_disabled(self) -> bool:
        """
        This method returns True if the element is visible.

        :return: returns True if the element does exist otherwise false.
        """
        return hasattr(self.raw_element, 'is_disabled') and self.raw_element.is_disabled

    def get_text_content(self) -> str:
        """
        This method returns the text of the element as a string.

        :return: returns the text of the element as a string
        """
        elem = self.raw_element
        if isinstance(elem, (Label, TextArea)):
            return elem.text
        if isinstance(elem, Button):
            return elem.label
        if isinstance(elem, (Input, Select, Digits)):
            return elem.value
        raise NotImplementedError(f'this method is not supported for widgets of type {elem}')

    def is_clickable(self) -> bool:
        """
        This method returns True if the element is (theoretically) clickable.

        :return: returns True if the element is (theoretically) clickable, otherwise False.
        """
        elem = self.raw_element
        if not isinstance(elem, (Button, Label)):
            return False
        return not self.is_visible() and elem.can_focus and hasattr(elem, 'on_click')

    ########################################################################################################
    # STATE
    ########################################################################################################

    def is_selected(self) -> bool:
        """
        The Is Element Selected command determines if the referenced element is selected or not. This operation only
        makes sense on input elements of the Checkbox- and Radio Button states, or on option elements.

        :return: returns True if the element is selected.
        """
        elem = self.raw_element
        if not isinstance(elem, Select):
            raise NotImplementedError(f'this method is not supported for widgets of type {self.raw_element}')
        return elem.selection is not None

    def get_rect(self) -> tuple[int, int, int, int]:
        """
        :return: the rect of the element (x, y, width, height)
        """
        elem = self.raw_element
        if isinstance(elem, Widget):
            return elem.region.x, elem.region.y, elem.region.width, elem.region.height
        raise NotImplementedError(f'this method is not supported for elements of type {self.raw_element}')

    ########################################################################################################
    # INTERACTIONS
    ########################################################################################################

    def click(self) -> None:
        """
        This method clicks the element.

        The Element Click command scrolls into view the element if it is not already pointer-interactable, and clicks
        its in-view center point.

        If the element's center point is obscured by another element, an element click intercepted error is returned.
        If the element is outside the viewport, an element not interactable error is returned.
        """
        self.driver.click(self.raw_element)

    def clear(self):
        """
        Clears the element
        """
        elem = self.raw_element
        if isinstance(elem, (Input, TextArea)):
            elem.clear()
        raise NotImplementedError(f'this method is not supported for widgets of type {self.raw_element}')

    def send_keys(self, text: str) -> None:
        """
        This method inserts a text into the field.

        The Element Send Keys command scrolls into view the form control element and then sends the provided keys to
        the element. In case the element is not keyboard-interactable, an element not interactable error is returned.

        :param text: the text that should be inserted into the field
        """
        self.driver.press(*text)

    def select_by_text(self, text_of_option_to_select: str) -> None:
        """
        This method selects the element by shown text.
        :param text_of_option_to_select: the expected text in the option that should be selected
        """
        elem = self.raw_element
        if not isinstance(elem, Select):
            raise NotImplementedError(f'this method is not supported for widgets of type {self.raw_element}')

        for label, value in elem.options:
            if label == text_of_option_to_select:
                elem.value = value
                break
        else:
            raise ValueError(f'can not find an option with the provided text `{text_of_option_to_select}`')

    def select_by_option_index(self, index: int) -> None:
        """
        This method selects the element by index.
        :param index: the option index that should be selected
        """
        elem = self.raw_element
        if not isinstance(elem, Select):
            raise NotImplementedError(f'this method is not supported for widgets of type {self.raw_element}')

        elem.value = elem.options[index][1]

    def select_by_option_value(self, value: str):
        """
        This method selects the element by value.
        :param value: the option value that should be selected
        """
        elem = self.raw_element
        if not isinstance(elem, Select):
            raise NotImplementedError(f'this method is not supported for widgets of type {self.raw_element}')
        elem.value = value

    def scroll_to_beginning(self):
        """
        This method scrolls to the beginning of the scrollable element. It needs to make sure, that after calling it,
        the scrollable element is always at the beginning of its content. It raises an error, in case the element can
        not be scrolled.
        """
        elem = self.raw_element
        if isinstance(elem, ScrollableContainer):
            elem.scroll_home(animate=False)
        raise ValueError(f'this method is only supported for widgets of class {ScrollableContainer.__name__}')

    def scroll_for(self, scroll_steps: int):
        """
        This method scrolls for one row / column.

        :param scroll_steps: the number of steps to scroll (positive: scrolls forward, negative: scrolls backward)
        """
        elem = self.raw_element

        if isinstance(elem, VerticalScroll):
            elem.scroll_relative(y=scroll_steps)
        elif isinstance(elem, HorizontalScroll):
            elem.scroll_relative(x=scroll_steps)
        else:
            raise ValueError(f'this method is only supported for widgets of class {VerticalScroll.__name__} or '
                             f'{HorizontalScroll.__name__}')

    def scroll_to_end(self):
        """
        This method scrolls to the end of the scrollable element.  It needs to make sure, that after calling it, the
        scrollable element is always at the end of its content. It raises an error, in case the element can not
        be scrolled.
        """
        elem = self.raw_element
        if isinstance(elem, ScrollableContainer):
            elem.scroll_home(animate=False)
        raise ValueError(f'this method is only supported for widgets of class {ScrollableContainer.__name__}')



import time
from typing import Literal

from balderhub.textual.lib.scenario_features import TextualControlFeature
from .. import scenario_features
from ..pages.calculator_page import CalculatorPage


class CalculatorKeysFeature(scenario_features.CalculatorFeature):

    page_calculator = CalculatorPage()
    textual_control = TextualControlFeature()

    def press_buttons(self, numbers: list[str]):
        self.textual_control.driver.press(*numbers)

    def read_result(self) -> float:
        return float(self.page_calculator.numbers.raw_element.value)

    def do_calc(self, first_number: float, second_number: float, operator: Literal["+", "-", "*", "/"]) -> float:
        self.press_buttons([e for e in str(first_number)])
        self.press_buttons([operator])
        self.press_buttons([e for e in str(second_number)])
        self.press_buttons(["="])
        return self.read_result()



from balderhub.gui.lib.utils.mixins import ClickableMixin, RectangleMixin

from ..textual_element import TextualElement


class Button(TextualElement, ClickableMixin, RectangleMixin):
    """
    A textual Button object (see: https://textual.textualize.io/widgets/button/)
    """
    def is_clickable(self) -> bool:
        return self._bridge.is_clickable()

    def click(self) -> None:
        self.bridge.click()

    @property
    def width(self) -> float:
        return self.bridge.get_rect()[2]

    @property
    def height(self) -> float:
        return self.bridge.get_rect()[3]



from balderhub.textual.lib.scenario_features.textual_page import TextualPage
from balderhub.textual.lib.utils import components
from balderhub.textual.lib.utils.selector import Selector


class CalculatorPage(TextualPage):

    @property
    def numbers(self) -> components.widgets.Digits:
        return components.widgets.Digits.by_selector(self.driver, Selector.by_id('numbers'))

    @property
    def btn_1(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-1'))

    @property
    def btn_2(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-2'))

    @property
    def btn_3(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-3'))

    @property
    def btn_4(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-4'))

    @property
    def btn_5(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-5'))

    @property
    def btn_6(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-6'))

    @property
    def btn_7(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-7'))

    @property
    def btn_8(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-8'))

    @property
    def btn_9(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('number-9'))

    @property
    def btn_ac(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('ac'))

    @property
    def btn_c(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('c'))

    @property
    def btn_plus_minus(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('plus-minus'))

    @property
    def btn_percent(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('percent'))

    @property
    def btn_divide(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('divide'))

    @property
    def btn_multiply(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('multiply'))

    @property
    def btn_minus(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('minus'))

    @property
    def btn_plus(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('plus'))

    @property
    def btn_point(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('point'))

    @property
    def btn_equals(self) -> components.widgets.Button:
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('equals'))




import balder

from tests.calculator.lib.scenario_features import CalculatorFeature


class ScenarioCalculator(balder.Scenario):

    class Calculator(balder.Device):
        calc = CalculatorFeature()

    @balder.parametrize('data_tuple', [
        (1, 2, 3),
        (-4.2, 2.2, -2),
        (1.111111111111111, 1.111111111111111, 2.222222222222222),
        (4, 2.8, 6.8),
        (1, 1, 2),
    ])
    def test_add(self, data_tuple: tuple):
        first_no, second_no, expected_result = data_tuple

        result = self.Calculator.calc.do_calc(first_no, second_no, '+')
        assert result == expected_result, f"got different result: {result} (expected {expected_result})"



from typing import Optional

import asyncio
import textual

from textual.dom import DOMNode
from textual.pilot import Pilot
from textual.widget import Widget

from balderhub.guicontrol.lib.utils.driver import BaseDriverClass
from balderhub.textual.lib.utils.selector import Selector
from balderhub.textual.lib.utils.element_bridges import \
    TextualElementBridge, TextualNotReidentifiableElementBridge, TextualFullyReidentifiableElementBridge



class TextualDriverClass(BaseDriverClass):
    """main textual driver class"""

    def __init__(self, app_instance: textual.app.App):
        self._app_instance = app_instance

        self._async_runner_cm = asyncio.Runner()
        self._async_runner = self._async_runner_cm.__enter__()

        self._async_pilot_cm = self._app_instance.run_test()

        self._pilot_instance = self._enter_pilot()

    def _enter_pilot(self) -> Pilot:

        async def run_query():
            await asyncio.sleep(0)
            pilot = await self._async_pilot_cm.__aenter__() # pylint: disable=unnecessary-dunder-call
            await asyncio.sleep(0)
            return pilot

        return self._async_runner.run(run_query())

    def get_bridge_for_raw_element(
            self,
            raw_element: DOMNode,
            parent: Optional[TextualElementBridge] = None
    ) -> TextualNotReidentifiableElementBridge:
        """
        This method returns the bridge object for a raw textual element.

        :param raw_element: the raw textual element
        :param parent: the parent bridge of the element that is a parent of this element (or None if there is no parent
                       specified)
        :return: the newly created bridge object
        """
        return TextualNotReidentifiableElementBridge(self, raw_element, parent=parent)

    def find_raw_element(self, selector: Selector) -> DOMNode:
        """
        This method returns the raw textual element that matches the given selector.

        :param selector: the selector to match
        """
        return self._app_instance.query_one(selector.to_textual_string())

    def find_raw_elements(self, selector: Selector) -> list[DOMNode]:
        """
        This method returns all matching raw textual elements that match the given selector.

        :param selector: the selector to match
        """
        return list(self._app_instance.query(selector.to_textual_string()))

    def find_bridge(self, selector: Selector) -> TextualFullyReidentifiableElementBridge:
        return TextualFullyReidentifiableElementBridge(self, selector, parent=None)

    def find_bridges(self, selector: Selector) -> list[TextualNotReidentifiableElementBridge]:
        items = []
        for cur_web_element in self.find_raw_elements(selector.to_textual_string()):
            items.append(self.get_bridge_for_raw_element(raw_element=cur_web_element, parent=None))
        return items

    def _shutdown_pilot(self) -> None:
        exc = (None, None, None)

        async def execute_shutdown():
            await asyncio.sleep(0)
            await self._async_pilot_cm.__aexit__(*exc)

        try:
            self._async_runner.run(execute_shutdown())
        finally:
            self._async_runner_cm.__exit__(*exc)
            self._async_runner = None
            self._async_runner_cm = None

            self._async_pilot_cm = None
            self._pilot_instance = None

    def quit(self):
        if self._pilot_instance is not None:
            self._shutdown_pilot()

    def click(self, widget: Widget):
        """
        Executes a mouse click on the given widget
        """
        return self._async_runner.run(self._pilot_instance.click(widget))

    def press(self, *keys: str):
        """
        This method executes key presses within the application.

        :param keys: the keys to press
        """
        return self._async_runner.run(self._pilot_instance.press(*keys))



from balderhub.gui.lib.utils.mixins import HasValueMixin

from ..textual_element import TextualElement


class Digits(TextualElement, HasValueMixin):
    """
    A textual Digits object (see: https://textual.textualize.io/widgets/digits/)
    """

    def get_value(self) -> str:
        return self.bridge.get_text_content()



from .collapsible import Collapsible
from .content_switcher import ContentSwitcher
from .list_view import ListView
from .radio_set import RadioSet
from .selection_list import SelectionList
from .tabbed_content import TabbedContent

__all__ = [
    "Collapsible",
    "ContentSwitcher",
    "ListView",
    "RadioSet",
    "SelectionList",
    "TabbedContent",
]



from balderhub.textual.lib.scenario_features import TextualPage
from balderhub.textual.lib.utils import components
from balderhub.textual.lib.utils.selector import Selector


class StopwatchPage(TextualPage):
    """
    This page represents the stopwatch screen in your Textual app.
    It defines properties for key widgets like the time display and buttons.
    """

    @property
    def numbers(self) -> components.widgets.Digits:
        """
        Returns the time display widget.
        Assumes it's a Digits widget with a tag 'TimeDisplay'.
        """
        return components.widgets.Digits.by_selector(self.driver, Selector.by_tag('TimeDisplay'))

    @property
    def btn_start(self) -> components.widgets.Button:
        """
        Returns the 'Start' button by its ID.
        """
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('start'))

    @property
    def btn_stop(self) -> components.widgets.Button:
        """
        Returns the 'Stop' button by its ID.
        """
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('stop'))

    @property
    def btn_reset(self) -> components.widgets.Button:
        """
        Returns the 'Reset' button by its ID.
        """
        return components.widgets.Button.by_selector(self.driver, Selector.by_id('reset'))


import datetime
import math
import time

import balder

from balderhub.textual.lib.scenario_features import TextualControlFeature

from .pages import StopwatchPage


class ScenarioStopWatch(balder.Scenario):

    class App(balder.Device):
        pass

    @balder.connect(App, over_connection=balder.Connection)
    class Controller(balder.Device):
        textual = TextualControlFeature()
        page = StopwatchPage()

    def test_start_stop(self):
        start_time = time.time()
        self.Controller.page.btn_start.click()
        time.sleep(1)
        self.Controller.page.btn_stop.click()
        expected_time = time.time() - start_time

        displayed_time = datetime.time.fromisoformat(self.Controller.page.numbers.get_value())

        displayed_sec = displayed_time.second + displayed_time.microsecond / 1_000_000

        assert displayed_time.hour == 0
        assert displayed_time.minute == 0
        # we are using this high deviation because of performance issues in Textual Pilot
        assert math.isclose(displayed_sec, expected_time, rel_tol=0.5), f"wrong time displayed: {displayed_sec} instead of {expected_time}"



from __future__ import annotations

from typing import TYPE_CHECKING, Union, Optional

from textual.dom import DOMNode

from ..selector import Selector
from .textual_element_bridge import TextualElementBridge


if TYPE_CHECKING:
    from ..drivers.textual_driver_class import TextualDriverClass
    from .textual_partly_reidentifiable_element_bridge import TextualPartlyReidentifiableElementBridge


class TextualNotReidentifiableElementBridge(TextualElementBridge):
    """
    A non reidentifiable element bridge can not re-identify an element by itself, because it does not specify absolute
    selectors.
    """

    def __init__(
            self,
            driver: TextualDriverClass,
            raw_element: DOMNode,
            parent: Optional[TextualElementBridge] = None
    ):
        super().__init__(driver=driver, parent=parent)
        self._raw_element = raw_element

    def __eq__(self, other):
        raise TypeError('can not compare elements that were created out of web elements')

    def find_bridge(self, selector: Selector) -> TextualPartlyReidentifiableElementBridge:
        # pylint: disable-next=import-outside-toplevel
        from .textual_partly_reidentifiable_element_bridge import TextualPartlyReidentifiableElementBridge

        return TextualPartlyReidentifiableElementBridge(self.driver, selector, parent=self)

    @property
    def parent(self) -> Union[TextualElementBridge, None]:
        return self._parent

    def exists(self):
        return self.raw_element.display



import balder
from balderhub.guicontrol.lib.scenario_features import GuiControlFeature

from balderhub.textual.lib.scenario_features.app_feature import AppFeature
from balderhub.textual.lib.utils.drivers.textual_driver_class import TextualDriverClass


class TextualControlFeature(GuiControlFeature):
    """
    Feature to control a textual application - necessary to interact with a textual application.
    """

    class App(balder.VDevice):
        """the app vdevice that provides the app instance under test"""
        app = AppFeature()

    @property
    def driver(self) -> TextualDriverClass:
        return super().driver

    def create(self) -> None:
        """
        creates the driver class to control the app - needs to be called within a fixture
        """
        self._driver = TextualDriverClass(self.App.app.get_app())



from . import containers

from .button import Button
from .checkbox import Checkbox
from .data_table import DataTable
from .digits import Digits
from .directory_tree import DirectoryTree
from .footer import Footer
from .header import Header
from .input import Input
from .label import Label
from .list_item import ListItem
from .loading_indicator import LoadingIndicator
from .log import Log
from .option_list import OptionList
from .placeholder import Placeholder
from .progress_bar import ProgressBar
from .radio_button import RadioButton
from .rule import Rule
from .select import Select
from .static import Static
from .switch import Switch
from .tabs import Tabs
from .text_area import TextArea
from .tree import Tree

__all__ = [
    "Button",
    "Checkbox",
    "DataTable",
    "Digits",
    "DirectoryTree",
    "Footer",
    "Header",
    "Input",
    "Label",
    "ListItem",
    "LoadingIndicator",
    "Log",
    "OptionList",
    "Placeholder",
    "ProgressBar",
    "RadioButton",
    "Rule",
    "Select",
    "Static",
    "Switch",
    "Tabs",
    "TextArea",
    "Tree",

]



from __future__ import annotations

from typing import Union, TYPE_CHECKING

from ..selector import Selector
from .textual_element_bridge import TextualElementBridge


if TYPE_CHECKING:
    from ..drivers.textual_driver_class import TextualDriverClass
    from ..components.textual_element import TextualElement
    from .textual_not_reidentifiable_element_bridge import TextualNotReidentifiableElementBridge


class TextualPartlyReidentifiableElementBridge(TextualElementBridge):
    """
    A partly reidentifiable element bridge specifies a bridge that does only have a relative selector to another bridge
    object that is not re-identifiable. This type of bridges are only reidentifiable if their parent element is in a
    reliable state.
    """
    def __init__(
            self,
            driver: TextualDriverClass,
            relative_selector: Selector,
            parent: Union[TextualPartlyReidentifiableElementBridge, TextualNotReidentifiableElementBridge]
    ):
        super().__init__(driver=driver, parent=parent)
        self._relative_selector = relative_selector

    def __eq__(self, other):
        if not isinstance(other, TextualPartlyReidentifiableElementBridge):
            raise TypeError(f'can not compare elements from different bridge type (this is {self.__class__} '
                            f'| other is {other.__class__})')
        if self.parent != other.parent:
            return False
        if self.relative_selector != other.relative_selector:
            return False
        return True

    @property
    def raw_element(self) -> TextualElement:
        if self._raw_element is None:
            self.re_identify_raw_element()
        return self._raw_element

    @property
    def parent(self) -> Union[TextualPartlyReidentifiableElementBridge, TextualNotReidentifiableElementBridge, None]:
        return super().parent

    def re_identify_raw_element(self) -> TextualElement:
        """
        This method re-identifies the element by requesting it again from the main driver. This method automatically
        updates the internal reference for this object.

        :return: the re-identified raw element
        """
        self._raw_element = self.parent.find_raw_element(self.relative_selector)
        return self._raw_element

    @property
    def relative_selector(self) -> Selector:
        """
        :return: returns the relative selector.
        """
        return self._relative_selector

    def find_bridge(self, selector: Selector) -> TextualPartlyReidentifiableElementBridge:
        return TextualPartlyReidentifiableElementBridge(self.driver, selector, parent=self)


    def exists(self) -> bool:
        if self.parent.find_raw_element(self.relative_selector):
            return True
        return False



import balder

from balderhub.textual.lib.scenario_features.textual_control_feature import TextualControlFeature
from tests.calculator.lib.pages.calculator_page import CalculatorPage
from tests.calculator.lib.scenario_features.calculator_app_feature import CalculatorAppFeature
from tests.calculator.lib.setup_features import CalculatorBtnFeature, CalculatorKeysFeature


class SetupCalculator(balder.Setup):

    class App(balder.Device):
        app = CalculatorAppFeature()

    class ControllerOverButtons(balder.Device):
        textual = TextualControlFeature(App="App")
        page_calc = CalculatorPage()
        calc = CalculatorBtnFeature()

    class ControllerOverKeyPresses(balder.Device):
        textual = TextualControlFeature(App="App")
        page_calc = CalculatorPage()
        calc = CalculatorKeysFeature()

    @balder.fixture('testcase')
    def setup_app_btn(self):
        self.ControllerOverButtons.textual.create()
        yield
        self.ControllerOverButtons.textual.quit()

    @balder.fixture('testcase')
    def setup_app_keys(self):
        self.ControllerOverKeyPresses.textual.create()
        yield
        self.ControllerOverKeyPresses.textual.quit()


import balder

from .dut.stopwatch import StopwatchApp

from balderhub.textual.lib.scenario_features import TextualControlFeature, AppFeature
from .pages import StopwatchPage


class MyAppFeature(AppFeature):

    def get_app(self):
        return StopwatchApp()


class SetupStopwatch(balder.Setup):

    class App(balder.Device):
        app = MyAppFeature()

    @balder.connect(App, over_connection=balder.Connection)
    class Controller(balder.Device):
        textual = TextualControlFeature(App="App")
        page = StopwatchPage()

    @balder.fixture('testcase')
    def setup_app(self):
        self.Controller.textual.create()
        yield
        self.Controller.textual.quit()



from ..textual_element import TextualElement


class LoadingIndicator(TextualElement):
    """
    A textual LoadingIndicator object (see: https://textual.textualize.io/widgets/loading_indicator/)
    """



import time
from typing import Literal

from balderhub.textual.lib.scenario_features import TextualControlFeature
from .. import scenario_features
from ..pages.calculator_page import CalculatorPage


class CalculatorBtnFeature(scenario_features.CalculatorFeature):

    page_calculator = CalculatorPage()
    textual_control = TextualControlFeature()

    def press_buttons(self, numbers: list[str]):
        mapping = {
            "1": self.page_calculator.btn_1,
            "2": self.page_calculator.btn_2,
            "3": self.page_calculator.btn_3,
            "4": self.page_calculator.btn_4,
            "5": self.page_calculator.btn_5,
            "6": self.page_calculator.btn_6,
            "7": self.page_calculator.btn_7,
            "8": self.page_calculator.btn_8,
            "9": self.page_calculator.btn_9,
            ".": self.page_calculator.btn_point,
            "+": self.page_calculator.btn_plus,
            "-": self.page_calculator.btn_minus,
            "*": self.page_calculator.btn_multiply,
            "/": self.page_calculator.btn_divide,
            "=": self.page_calculator.btn_equals,
        }
        for cur_char in numbers:
            if cur_char not in mapping.keys():
                raise KeyError(f"Unexpected character {cur_char}")
            mapping[cur_char].click()
            time.sleep(0.3)

    def read_result(self) -> float:
        return float(self.page_calculator.numbers.raw_element.value)

    def do_calc(self, first_number: float, second_number: float, operator: Literal["+", "-", "*", "/"]) -> float:
        self.press_buttons([e for e in str(first_number)])
        self.press_buttons([operator])
        self.press_buttons([e for e in str(second_number)])
        self.press_buttons(["="])
        return self.read_result()



from .textual_element_bridge import TextualElementBridge
from .textual_fully_reidentifiable_element_bridge import TextualFullyReidentifiableElementBridge
from .textual_not_reidentifiable_element_bridge import TextualNotReidentifiableElementBridge
from .textual_partly_reidentifiable_element_bridge import TextualPartlyReidentifiableElementBridge


__all__ = [
    'TextualElementBridge',
    'TextualFullyReidentifiableElementBridge',
    'TextualNotReidentifiableElementBridge',
    'TextualPartlyReidentifiableElementBridge',
]



from enum import Enum

from balderhub.gui.lib.utils.base_selector import BaseSelector


class Selector(BaseSelector):
    """specific textual selector class"""

    class By(Enum):
        """
        Enum that describes the selecting type
        """
        ID = 'id'
        CLASSNAME = 'class name'
        TAG = "tag name"

    @classmethod
    def by_id(cls, name: str):
        """creates a new selector identified by the id"""
        return cls(cls.By.ID, name)

    @classmethod
    def by_tag(cls, name: str):
        """creates a new selector identified by the tag name"""
        return cls(cls.By.TAG, name)

    @classmethod
    def by_class(cls, name: str):
        """creates a new selector identified by a class name"""
        return cls(cls.By.CLASSNAME, name)

    def to_textual_string(self):
        """
        Converts the selector into the textual notation
        """
        if self._by_type == self.By.ID:
            return f'#{self._identifier}'
        if self._by_type == self.By.CLASSNAME:
            return f'.{self._identifier}'
        if self._by_type == self.By.TAG:
            return f'{self._identifier}'
        raise KeyError(f"Unknown selector type: {self._by_type}")



from __future__ import annotations

from typing import Union

from textual.dom import DOMNode

from balderhub.gui.lib.utils.components.abstract_element import AbstractElement
from balderhub.gui.lib.utils.mixins import VisibleMixin
from balderhub.textual.lib.utils.drivers import TextualDriverClass
from balderhub.textual.lib.utils.element_bridges import TextualElementBridge

from ..selector import Selector


class TextualElement(AbstractElement, VisibleMixin):
    """
    The basic Textual Element
    """

    def __init__(self, bridge: TextualElementBridge):
        self._bridge = bridge

    def __eq__(self, other):
        return self.bridge == other.bridge

    @classmethod
    def by_selector(
            cls,
            driver: TextualDriverClass,
            selector: Selector,
            parent: Union[TextualElement, TextualElementBridge] = None
    ):
        """
        This method returns the element for the provided selector.

        :param driver: the textual driver
        :param selector: the selector to identify the element
        :param parent: optional a parent textual element, if the selector is relative
        :return: the textual element that is identified by the selector
        """
        if parent is None:
            bridge = driver.find_bridge(selector)
        else:
            if isinstance(parent, TextualElement):
                parent = parent.bridge
            bridge = parent.find_bridge(selector)
        return cls(bridge)

    @classmethod
    def by_raw_element(
            cls,
            driver: TextualDriverClass,
            raw_element: DOMNode,
            parent: Union[TextualElement, TextualElementBridge] = None
    ):
        """
        This method returns the textual element by the raw textual framework element.

        :param driver: the textual driver
        :param raw_element: the raw textual engine element (f.e. ``DOMNode`` or ``Widget``)
        :param parent: optional a parent textual element, if the selector is relative
        :return: the textual element that is identified by the selector
        """
        if parent is not None and isinstance(parent, TextualElement):
            parent = parent.bridge
        bridge = driver.get_bridge_for_raw_element(raw_element, parent)
        return cls(bridge)

    @property
    def bridge(self) -> TextualElementBridge:
        """
        :return: returns the underlying bridge object
        """
        return self._bridge

    @property
    def driver(self) -> TextualDriverClass:
        """
        :return: returns the underlying textual driver
        """
        return self._bridge.driver

    @property
    def raw_element(self) -> DOMNode:
        """
        :return: returns the raw textual engine element
        """
        return self._bridge.raw_element

    @property
    def parent_bridge(self) -> TextualElementBridge:
        """
        :return: returns the bridge of the parent element (if any)
        """
        return self._bridge.parent

    def exists(self) -> bool:
        return self._bridge.exists()

    def is_visible(self) -> bool:
        return self._bridge.is_visible()



import textual.app
from balder import Feature


class AppFeature(Feature):
    """
    Basic feature that provides the textual app instance that should be tested.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._async_pilot_cm = None
        self._pilot_instance = None

    def get_app(self) -> textual.app.App:
        """
        Returns the app instance that should be tested.
        """
        raise NotImplementedError()



from ...textual_element import TextualElement


class ContentSwitcher(TextualElement):
    """
    A textual ContentSwitcher object (see: https://textual.textualize.io/widgets/content_switcher/)
    """



from __future__ import annotations


from typing import TYPE_CHECKING, Optional, Union

from textual.dom import DOMNode

from ..selector import Selector
from .textual_element_bridge import TextualElementBridge


if TYPE_CHECKING:
    from ..drivers.textual_driver_class import TextualDriverClass
    from ..components.textual_element import TextualElement


class TextualFullyReidentifiableElementBridge(TextualElementBridge):
    """
    A fully reidentifiable element bridge specifies the element in an absolute manner. It doesn't matter if the element
    has a full absolute selector or has parents with re-identifiable selectors (so all parents need to be
    :class:`FullyReidentifiableElementBridge` objects)
    """

    def __init__(
            self,
            driver: TextualDriverClass,
            selector: Selector,
            parent: Optional[TextualFullyReidentifiableElementBridge] = None
    ):
        super().__init__(driver=driver, parent=parent)
        self._selector = selector

    def __eq__(self, other):
        if not isinstance(other, TextualFullyReidentifiableElementBridge):
            raise TypeError(f'can not compare elements from different bridge type (this is {self.__class__} '
                            f'| other is {other.__class__})')
        if self.parent != other.parent:
            return False
        if self.selector != other.selector:
            return False
        return True

    @property
    def raw_element(self) -> TextualElement:
        if self._raw_element is None:
            self.re_identify_raw_element()
        return self._raw_element

    @property
    def parent(self) -> Union[TextualFullyReidentifiableElementBridge, None]:
        return super().parent

    @property
    def selector(self) -> Selector:
        """
        :return: the absolute selector this element bridge has.
        """
        return self._selector

    def re_identify_raw_element(self) -> DOMNode:
        """
        This method re-identifies the element by requesting it again from the main driver. This method automatically
        updates the internal reference for this object.

        :return: the re-identified raw element
        """
        if self.parent is None:
            self._raw_element = self.driver.find_raw_element(self.selector)
        else:
            self._raw_element = self.parent.find_raw_element(self.selector)
        return self._raw_element

    def find_bridge(self, selector: Selector) -> TextualFullyReidentifiableElementBridge:
        return TextualFullyReidentifiableElementBridge(self.driver, selector, parent=self)

    def exists(self) -> bool:
        if self.parent is None:
            if self.driver.find_raw_element(self.selector):
                return True
        else:
            if self.parent.find_raw_element(self.selector):
                return True

        return False
